"use strict";
/**
 * Copyright (c) 2024 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author David Sehnal <david.sehnal@gmail.com>
 * @author Cai Huiyu <szmun.caihy@gmail.com>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ExternalVolumeColorThemeProvider = exports.ExternalVolumeColorTheme = exports.ExternalVolumeColorThemeParams = void 0;
const color_1 = require("../../mol-util/color");
const param_definition_1 = require("../../mol-util/param-definition");
const volume_1 = require("../../mol-model/volume");
const location_iterator_1 = require("../../mol-geo/util/location-iterator");
const linear_algebra_1 = require("../../mol-math/linear-algebra");
const interpolate_1 = require("../../mol-math/interpolate");
const categories_1 = require("./categories");
const Description = `Assigns a color based on volume value at a given vertex.`;
exports.ExternalVolumeColorThemeParams = {
    volume: param_definition_1.ParamDefinition.ValueRef((ctx) => {
        const volumes = ctx.state.data.selectQ(q => q.root.subtree().filter(c => { var _a; return volume_1.Volume.is((_a = c.obj) === null || _a === void 0 ? void 0 : _a.data); }));
        return volumes.map(v => { var _a, _b; return [v.transform.ref, (_b = (_a = v.obj) === null || _a === void 0 ? void 0 : _a.label) !== null && _b !== void 0 ? _b : '<unknown>']; });
    }, (ref, getData) => getData(ref)),
    coloring: param_definition_1.ParamDefinition.MappedStatic('absolute-value', {
        'absolute-value': param_definition_1.ParamDefinition.Group({
            domain: param_definition_1.ParamDefinition.MappedStatic('auto', {
                custom: param_definition_1.ParamDefinition.Interval([-1, 1]),
                auto: param_definition_1.ParamDefinition.Group({
                    symmetric: param_definition_1.ParamDefinition.Boolean(false, { description: 'If true the automatic range is determined as [-|max|, |max|].' })
                })
            }),
            list: param_definition_1.ParamDefinition.ColorList('red-white-blue', { presetKind: 'scale' })
        }),
        'relative-value': param_definition_1.ParamDefinition.Group({
            domain: param_definition_1.ParamDefinition.MappedStatic('auto', {
                custom: param_definition_1.ParamDefinition.Interval([-1, 1]),
                auto: param_definition_1.ParamDefinition.Group({
                    symmetric: param_definition_1.ParamDefinition.Boolean(false, { description: 'If true the automatic range is determined as [-|max|, |max|].' })
                })
            }),
            list: param_definition_1.ParamDefinition.ColorList('red-white-blue', { presetKind: 'scale' })
        })
    }),
    defaultColor: param_definition_1.ParamDefinition.Color((0, color_1.Color)(0xcccccc)),
    normalOffset: param_definition_1.ParamDefinition.Numeric(0., { min: 0, max: 20, step: 0.1 }, { description: 'Offset vertex position along its normal by given amount.' }),
};
function ExternalVolumeColorTheme(ctx, props) {
    let volume;
    try {
        volume = props.volume.getValue();
    }
    catch (_a) {
        // .getValue() is resolved during state reconciliation => would throw from UI
    }
    // NOTE: this will currently be slow for with GPU/texture meshes due to slow iteration
    // TODO: create texture to be able to do the sampling on the GPU
    let color;
    if (volume) {
        const coloring = props.coloring.params;
        const { stats } = volume.grid;
        const domain = coloring.domain.name === 'custom' ? coloring.domain.params : [stats.min, stats.max];
        const isRelative = props.coloring.name === 'relative-value';
        if (coloring.domain.name === 'auto' && isRelative) {
            domain[0] = (domain[0] - stats.mean) / stats.sigma;
            domain[1] = (domain[1] - stats.mean) / stats.sigma;
        }
        if (props.coloring.params.domain.name === 'auto' && props.coloring.params.domain.params.symmetric) {
            const max = Math.max(Math.abs(domain[0]), Math.abs(domain[1]));
            domain[0] = -max;
            domain[1] = max;
        }
        const scale = color_1.ColorScale.create({ domain, listOrName: coloring.list.colors });
        const cartnToGrid = volume_1.Grid.getGridToCartesianTransform(volume.grid);
        linear_algebra_1.Mat4.invert(cartnToGrid, cartnToGrid);
        const gridCoords = (0, linear_algebra_1.Vec3)();
        const { dimensions, get } = volume.grid.cells.space;
        const data = volume.grid.cells.data;
        const [mi, mj, mk] = dimensions;
        color = (location) => {
            if (!(0, location_iterator_1.isPositionLocation)(location)) {
                return props.defaultColor;
            }
            // Offset the vertex position along its normal
            linear_algebra_1.Vec3.copy(gridCoords, location.position);
            if (props.normalOffset > 0) {
                linear_algebra_1.Vec3.scaleAndAdd(gridCoords, gridCoords, location.normal, props.normalOffset);
            }
            linear_algebra_1.Vec3.transformMat4(gridCoords, gridCoords, cartnToGrid);
            const i = Math.floor(gridCoords[0]);
            const j = Math.floor(gridCoords[1]);
            const k = Math.floor(gridCoords[2]);
            if (i < 0 || i >= mi || j < 0 || j >= mj || k < 0 || k >= mk) {
                return props.defaultColor;
            }
            const u = gridCoords[0] - i;
            const v = gridCoords[1] - j;
            const w = gridCoords[2] - k;
            // Tri-linear interpolation for the value
            const ii = Math.min(i + 1, mi - 1);
            const jj = Math.min(j + 1, mj - 1);
            const kk = Math.min(k + 1, mk - 1);
            let a = get(data, i, j, k);
            let b = get(data, ii, j, k);
            let c = get(data, i, jj, k);
            let d = get(data, ii, jj, k);
            const x = (0, interpolate_1.lerp)((0, interpolate_1.lerp)(a, b, u), (0, interpolate_1.lerp)(c, d, u), v);
            a = get(data, i, j, kk);
            b = get(data, ii, j, kk);
            c = get(data, i, jj, kk);
            d = get(data, ii, jj, kk);
            const y = (0, interpolate_1.lerp)((0, interpolate_1.lerp)(a, b, u), (0, interpolate_1.lerp)(c, d, u), v);
            let value = (0, interpolate_1.lerp)(x, y, w);
            if (isRelative) {
                value = (value - stats.mean) / stats.sigma;
            }
            return scale.color(value);
        };
    }
    else {
        color = () => props.defaultColor;
    }
    return {
        factory: ExternalVolumeColorTheme,
        granularity: 'vertex',
        preferSmoothing: true,
        color,
        props,
        description: Description,
        // TODO: figure out how to do legend for this
    };
}
exports.ExternalVolumeColorTheme = ExternalVolumeColorTheme;
exports.ExternalVolumeColorThemeProvider = {
    name: 'external-volume',
    label: 'External Volume',
    category: categories_1.ColorThemeCategory.Misc,
    factory: ExternalVolumeColorTheme,
    getParams: () => exports.ExternalVolumeColorThemeParams,
    defaultValues: param_definition_1.ParamDefinition.getDefaultValues(exports.ExternalVolumeColorThemeParams),
    isApplicable: (ctx) => true,
};
